// Copyright (C) 2022 - 2023, Xilinx Inc - All rights reserved
//
// Licensed under the Apache License, Version 2.0 (the "License"). You may
// not use this file except in compliance with the License. A copy of the
// License is located at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
// WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
// License for the specific language governing permissions and limitations
// under the License.

#include <cassert>
#include <cstring>
#include "xrm_enc_interface.h"
#include "xrm_utils.h"

/**
 * @brief Release the xrm resources which were allocated. Destroy the xrm
 * context.
 * 
 * @param xrm_enc_ctx: Encoder XRM context
 * @return None
 */
void xrm_enc_release(XrmEncodeContext* xrm_enc_ctx) {
  if (!xrm_enc_ctx) {
    return;
  }
  if (xrm_enc_ctx->enc_res_in_use) {
    xrmCuListReleaseV2(xrm_enc_ctx->xrm_ctx, &xrm_enc_ctx->encode_cu_list_res);
    xrm_enc_ctx->enc_res_in_use = 0;
  }
  xrmDestroyContext(xrm_enc_ctx->xrm_ctx);
}

/**
 * @brief Updates the group info for the encoder / lookahead slice
 *
 * @param enc_cu_pool_prop The pool property which contains the cu / group info
 * @param slice_id The id of the slice (-1 to let xrm decide, otherwise 0, 1)
 * @param is_ull Is ultra low latency enabled
 * @return int32_t XRM_SUCCESS on success or XRM_ERROR on error
 */
static int32_t enc_update_group_info(xrmCuPoolPropertyV2* enc_cu_pool_prop, int slice_id, bool is_ull, bool is_xav1) {
  xrmCuListPropertyV2* enc_list_props       = &enc_cu_pool_prop->cuListProp;
  int                  cu_num               = enc_list_props->cuNum;
  bool                 is_lookahead_enabled = cu_num > 1;
  uint32_t             constraint_info      = XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_PSEUDO_LOGIC_GROUP_ID;
  int32_t              group_id             = -1;

  if (slice_id != -1) {
    constraint_info = XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_REAL_LOGIC_GROUP_ID;
    switch (slice_id) {
    case 0:
      if (is_xav1) {
        group_id = XAV1_SLICE_0;
      } else {
        group_id = VC8000E_SLICE_0;
      }
      break;
    case 1:
      if (is_xav1) {
        group_id = XAV1_SLICE_1;
      } else {
        group_id = VC8000E_SLICE_1;
      }
      break;
    default:
      fprintf(stderr, "Invalid slice %d given to %s!\n", slice_id, __func__);
      return XRM_ERROR;
    }
  } else if (is_ull) {
    constraint_info = XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_REAL_LOGIC_GROUP_ID;
    if (is_xav1) {
      group_id = XAV1_SLICE_0;
    } else {
      group_id = VC8000E_SLICE_0;
    }
    enc_list_props->cuProps[0].requestLoad /= 2;
    if (is_lookahead_enabled) {
      enc_list_props->cuProps[1].requestLoad /= 2;
    }
  } else if (!is_lookahead_enabled) {
    constraint_info = XRM_LOGIC_GROUP_INFO_CONSTRAINT_TYPE_LOGIC_GROUP_TYPE;
  }

  xrmCuPropertyLogicGroupInfoV2* enc_logic_group_info = &enc_list_props->cuProps[0].logicGroupInfo;
  enc_logic_group_info->constraintInfo                = constraint_info;
  enc_logic_group_info->groupId                       = group_id;
  strcpy(enc_logic_group_info->groupType, is_xav1 ? "ENCODER_TYPE1" : "ENCODER_TYPE2");

  if (is_lookahead_enabled) {
    xrmCuPropertyLogicGroupInfoV2* la_logic_group_info = &enc_list_props->cuProps[1].logicGroupInfo;
    la_logic_group_info->constraintInfo                = constraint_info;
    la_logic_group_info->groupId                       = group_id;
  }

  if (is_ull) {
    // TODO: What if slice 1 is given and ULL is enabled? Should we error or put the load on the other slice too, or both on the same slice?
    if (is_xav1) {
      group_id = XAV1_SLICE_1;
    } else {
      group_id = VC8000E_SLICE_1;
    }
    memcpy(&enc_list_props->cuProps[cu_num], &enc_list_props->cuProps[0], sizeof(enc_list_props->cuProps[cu_num]));
    enc_list_props->cuProps[cu_num].logicGroupInfo.groupId = group_id;
    cu_num++;
    if (is_lookahead_enabled) {
      memcpy(&enc_list_props->cuProps[cu_num], &enc_list_props->cuProps[1], sizeof(enc_list_props->cuProps[cu_num]));
      enc_list_props->cuProps[cu_num].logicGroupInfo.groupId = group_id;
      cu_num++;
    }
    enc_cu_pool_prop->cuListProp.cuNum = cu_num;
  }
  return XRM_SUCCESS;
}

/**
 * @brief Check & allocate the necessary xrm resources given the relevant 
 * encoder parameters
 * 
 * @param xrm_enc_ctx Used to store information about resources this function
 * allocates so they can be safely released later
 * @param dev_index The device index
 * @param slice_id The slice index. -1 for xrm to choose the slice, otherwise
 * 0 or 1
 * @param is_xav1 Is the codec to be used xav1 or not
 * @param is_ull Is ultra low latency enabled or not
 * @param xrm_props Contains relevant information for load calculation
 * @return XRM_SUCCESS on success, XRM_ERROR on error or XRM_ERROR_CONNECT_FAIL
 * when unable to connect to daemon
 */
int32_t xrm_enc_reserve(XrmEncodeContext* xrm_enc_ctx, int dev_index, int slice_id, bool is_xav1, bool is_ull, XrmInterfaceProperties* xrm_props) {
  xrmCuPoolPropertyV2 enc_cu_pool_prop;
  int32_t             ret = XRM_ERROR;
  memset(&enc_cu_pool_prop, 0, sizeof(enc_cu_pool_prop));
  xrm_enc_ctx->dev_index = dev_index;
  xrm_enc_ctx->slice_id  = slice_id;

  /* Create XRM context */
  xrm_enc_ctx->xrm_ctx = xrmCreateContext(XRM_API_VERSION_1);
  if (xrm_enc_ctx->xrm_ctx == NULL) {
    return XRM_ERROR_CONNECT_FAIL;
  }

  /* Calculate encoder load based on encoder properties */
  std::string ip_name = "ENCODER";
  ret                 = xrm_utils_load_calc(xrm_enc_ctx->xrm_ctx, xrm_props, NULL, 0, ip_name, &xrm_enc_ctx->enc_load[0], is_xav1);
  if (ret <= XRM_ERROR) {
    fprintf(stderr, "Enc load calculation failed %d\n", ret);
    return ret;
  }

  if (xrm_enc_ctx->enc_load[1]) {
    xrm_utils_update_pool_props("ENCODER", xrm_enc_ctx->dev_index, xrm_enc_ctx->enc_load[1], &enc_cu_pool_prop, is_xav1);
  }
  if (xrm_enc_ctx->enc_load[2]) {
    xrm_utils_update_pool_props("LOOKAHEAD", xrm_enc_ctx->dev_index, xrm_enc_ctx->enc_load[2], &enc_cu_pool_prop, is_xav1);
  }

  (void) is_ull;
  if (enc_update_group_info(&enc_cu_pool_prop, slice_id, 0, is_xav1) <= XRM_ERROR) { // Ignore ULL for now
    return XRM_ERROR;
  }
  ret                         = xrm_utils_reserve_resources(xrm_enc_ctx->xrm_ctx, &enc_cu_pool_prop, &xrm_enc_ctx->encode_cu_list_res);
  xrm_enc_ctx->enc_res_in_use = ret == XRM_SUCCESS ? 1 : 0;
  xrm_enc_ctx->dev_index      = xrm_enc_ctx->encode_cu_list_res.cuResources[0].deviceId;

  switch (xrm_enc_ctx->encode_cu_list_res.cuResources[0].logicGroupInfor.groupId) {
  case VC8000E_SLICE_0:
  case XAV1_SLICE_0:
    xrm_enc_ctx->slice_id = 0;
    break;
  case VC8000E_SLICE_1:
  case XAV1_SLICE_1:
    xrm_enc_ctx->slice_id = 1;
    break;
  default:
    fprintf(stderr, "XRM calculated invalid slice %d!\n", xrm_enc_ctx->encode_cu_list_res.cuResources[0].logicGroupInfor.groupId);
    return XRM_ERROR;
  }
  return ret;
}
